---
title: Service Programming Model
type: Details
---

You can configure a microservice deployed in Kyma to receive Events from the Event Bus by creating a Subscription custom resource.

## Event Delivery

The Event is delivered as an `HTTP POST` request. Event Metadata is a part of an HTTP request Headers. Event Payload is the body of the request.

![TEST](./assets/service-programming-model.png)

The Event Delivery workflow is as follows:

1. The Event is published to the Kyma Event Bus from an external system instance in a bound Application.
2. The Event Bus checks for the Event subscription and activation. It creates an `HTTP POST` request using Event Payload and Metadata.
3. The Service receives the `HTTP POST` request. The Event Metadata is represented in the `HTTP Request Headers` request and the Event Payload is represented in the `HTTP Request Body`.

## Event Metadata

The following HTTP Headers provide information about the Event Metadata.

|Header| Description|
|------|--------|
| **kyma-event-id** | Business Event's ID delivered to the microservice. |
| **kyma-event-time** | Business Event's time delivered to the microservice. |
| **kyma-event-type** | Business Event's type delivered to the microservice. |
| **kyma-event-type-version** | Business Event's version delivered to the microservice. |
| **kyma-source-environment** | Business Event's source environment delivered to the microservice. |
| **kyma-source-namespace** | Business Event's source Namespace delivered to the microservice. |
| **kyma-source-type** | Business Event's source type delivered to the microservice. |
| **kyma-subscription** | Subscription name defined in the subscription contract, or in a CRD. This business Event is published to its subscribers. |
| **x-b3-flags** | Header used by the Zipkin tracer in Envoy. It encodes one or more options. See more on Zipkin tracing [here](https://github.com/openzipkin/b3-propagation). |
| **x-b3-parentspanid** | Header used by the Zipkin tracer in Envoy. The **ParentSpanId** is 64-bit in length and indicates the position of the parent operation in the trace tree. When the span is the root of the trace tree, the **ParentSpanId** is absent. |
| **x-b3-sampled** | Header used by the Zipkin tracer in Envoy. When the **Sampled** flag is either not specified or set to `1`, the span is reported to the tracing system. Once **Sampled** is set to `0` or `1`, the same value should be consistently sent downstream. |
| **x-b3-spanid** | Header used by the Zipkin tracer in Envoy. The **SpanId** is 64-bit in length and indicates the position of the current operation in the trace tree. The value should not be interpreted. It may or may not be derived from the value of the **TraceId**. |
| **x-b3-traceid** | Header used by the Zipkin tracer in Envoy. The **TraceId** is 64-bit in length and indicates the overall ID of the trace. Every span in a trace shares this ID. |
| **x-request-id** | Randomly generated ID which identifies the HTTP request delivering the business Event. |
| **x-envoy-decorator-operation** | If this header is present in Ingress requests, its value overrides any locally defined operation (span) name on the server span generated by the tracing mechanism. If this header is present in an Egress response, its value overrides any locally defined operation (span) name on the client span. |
| **x-envoy-expected-rq-timeout-ms** | Time in milliseconds in which the router expects the request to be completed. Envoy sets this header so that the upstream host receiving the request can make decisions based on the request timeout. It is set on internal requests and is either taken from the **x-envoy-upstream-rq-timeout-ms** header or from the route timeout. |
| **x-istio-attributes** | Istio specific metadata. |

## Event Payload

The Event Payload is delivered as the body of the HTTP Request in JSON format. The JSON schema is available in the Service Catalog in the registered service for the remote Events.

## Event Payload Example

In this example, you write a service for an `order.created` Event published by the external solution service. The published Event schema looks as follows:

```json
{
  "example": {
    "orderCode": "4caad296-e0c5-491e-98ac-0ed118f9474e"
  },
  "properties": {
    "orderCode": {
      "description": "Resource identifier",
      "title": "OrderCode",
      "type": "string"
    }
  },
  "type": "object"
}
```

The HTTP POST request payload is a JSON object:

```json
{"orderCode": "4caad296-e0c5-491e-98ac-0ed118f9474e"}
```

## Successful Delivery

A message delivered to a subscriber is considered successfully consumed if the service's HTTP response status code is `2xx`. If the status code is not `2xx` (< 200 or >= 300), it means that a message consumption is not successful and that the message delivery is re-tried. This implies **At-least-once** delivery guarantee.

## Event Subscription Service Example

Refer to [this](https://github.com/kyma-project/examples/tree/master/event-subscription/service) example to find a complete scenario for implementing a subscriber service to a business Event.
